"""
Provide implementation of the BaseGenerator interfaces.
"""
from abc import (
    ABC,
    abstractmethod,
)


class BaseGeneratorInterface(ABC):
    """
    Implements BaseGenerator interface.
    """

    @abstractmethod
    def generate(self) -> str:
        """
        Generate `EOS` name.

        :return: `EOS` name
        """

    @abstractmethod
    def generate_list(self, num: int) -> list:
        """
        Generate list of `EOS` names.

        :param num: number of generated names in list.
        :return: `EOS` name
        """



"""Top-level module for `EOS` name generator.

This module generate random name which suits for `eosio blockchain` name
conversation.

The module provides 3 methods to generate a name:
    - random name generation based in a ready-made data (word dictionary)
    - markov chain text generation
    - recurrent neural network text generation
"""
from eos_name_generator.random_generator.generator import RandomNameGenerator

__version__ = "0.1.0"
__version_info__ = tuple(
    int(i) for i in __version__.split(".") if i.isdigit()
)



"""
Provide errors for BaseGenerator interface.
"""


class ValidationDataError(Exception):
    """
    Name generation data contains invalid characters or does not match the name length error.
    """

    def __init__(self, message):
        self.message = message



"""
Provide constants for BaseGenerator interface.
"""
from os.path import dirname

from eos_name_generator.utils import FastRandomChoice

EOS_NAME_LENGTH = 12
SEED_DATA_PATH = dirname(__file__) + '/' + 'random_generator/seed_data/nounlist.txt'
NUMBERS_PROBABILITIES = 0.1
RANDOM_PROVIDER_INSTANCE = FastRandomChoice()



"""
Provide an implementation of the RandomNameGenerator interface.
"""
from eos_name_generator.constants import (
    EOS_NAME_LENGTH,
    NUMBERS_PROBABILITIES,
    RANDOM_PROVIDER_INSTANCE,
    SEED_DATA_PATH,
)
from eos_name_generator.interfaces import BaseGeneratorInterface
from eos_name_generator.random_generator.data_reader import DataReader


class RandomNameGenerator(BaseGeneratorInterface):
    """
    Implementation of the RandomNameGenerator.
    """

    def __init__(
            self,
            seed_data_path=SEED_DATA_PATH,
            numbers_probabilities=NUMBERS_PROBABILITIES,
            random_provider_instance=RANDOM_PROVIDER_INSTANCE,
    ):
        """
        `RandomNameGenerator` constructor.

        :param seed_data_path: path to the data based on which the name will be generated.
        :param numbers_probabilities: the probability of occurrence of numbers in the generated word.
        :param random_provider_instance: the random provider instance.
        """
        self._seed_data_path = seed_data_path
        self.numbers_probabilities = numbers_probabilities
        self.random_provider = random_provider_instance
        self.data_provider = DataReader

    def generate(self) -> str:
        """
        Generate `EOS` name method.

        :return: `EOS` name str
        """
        base_dict_keys = list(self.__base_dict.keys())
        base_word_len = self.random_provider.choice(base_dict_keys, p=self.__probabilities_len_base_word)
        additional_word_len = EOS_NAME_LENGTH - base_word_len

        base_words = self.__base_dict.get(base_word_len)
        additional_alphabet_words = []
        if additional_word_len in self.__base_dict:
            additional_alphabet_words = self.__base_dict.get(additional_word_len)

        name = self.__get_random_name(base_words, additional_alphabet_words)

        return name

    def generate_list(self, num: int) -> list:
        """
        Generate list of `EOS` names method.

        :param num: number of generated names in list.
        :return: `EOS` name
        """
        generated_list = []
        for _ in range(num):
            name = self.generate()
            generated_list.append(name)

        return generated_list

    @property
    def seed_data_path(self) -> str:
        """
        Get `seed_data_path` variable.

        :return: `generation_base_data_path` string value
        """
        return self._seed_data_path

    @seed_data_path.setter
    def seed_data_path(self, value):
        """
        Set `seed_data_path` value and generate new basic dictionary and probabilities word length.

        :param value: `seed_data_path` variable value
        """
        self._seed_data_path = value
        self.data_provider.data_path = value
        self.__base_dict = self.data_provider.get_dictionary_by_word_len()
        self.__probabilities_len_base_word = self.__get_probabilities_len_base_word()

    @property
    def numbers_probabilities(self) -> int:
        """
        Get `numbers_probabilities` variable.

        :return: `numbers_probabilities` int value
        """
        return self._numbers_probabilities

    @numbers_probabilities.setter
    def numbers_probabilities(self, value):
        """
        Set `numbers_probabilities` value.

        :param value: `numbers_probabilities` variable value
        """
        if 0 < value > 1:
            raise ValueError("The numbers probabilities value must be between 1 and 0.")

        self._numbers_probabilities = value

    @property
    def random_provider(self):
        """
        Get `random_provider` variable.

        :return: `random_provider` instance.
        """
        return self._random_provider

    @random_provider.setter
    def random_provider(self, value):
        """
        Set `random_provider` value.

        :param value: `random_provider` instance.
        """
        random_provider_dir = dir(value)
        required_methods = ['choice', 'randint']

        for method_name in required_methods:
            if method_name not in random_provider_dir:

                error_message = f'The interface `random_provider` does not contain {method_name} method.'
                raise AttributeError(error_message)

        self._random_provider = value

    @property
    def data_provider(self):
        """
        Get `data_provider` variable.

        :return: `data_provider` instance.
        """
        return self._data_provider

    @data_provider.setter
    def data_provider(self, value):
        """
        Set `data_provider` variable.

        :param value: `data_provider` variable value.
        """
        data_provider_dir = dir(value)
        required_method = 'get_dictionary_by_word_len'

        if required_method not in data_provider_dir:

            error_message = f'The interface `data_provider` does not contain {required_method} method.'
            raise AttributeError(error_message)

        self._data_provider = value(self.seed_data_path)
        self.__base_dict = self.data_provider.get_dictionary_by_word_len()
        self.__probabilities_len_base_word = self.__get_probabilities_len_base_word()

    def __get_random_name(self, base_words, additional_alphabet_words) -> str:
        """
        Generate random name based on `base_words` and `additional_alphabet_words`.

        :param base_words: list of words on the basis of which the name will be created
        :param additional_alphabet_words: additional words to be added to the base word
        :return: random name string
        """
        base_word_random_index = self.random_provider.randint(0, len(base_words) - 1)
        base_word = base_words[base_word_random_index]
        additional_alphabet_words_len = len(additional_alphabet_words)
        additional_word_alphabet_probability = self.__get_probability_alphabet_additional_word(
            additional_alphabet_words_len,
        )

        numbers_probabilities = self.numbers_probabilities if additional_word_alphabet_probability else 1
        additional_words_probabilities = [additional_word_alphabet_probability, numbers_probabilities]
        is_additional_alphabet_word = self.random_provider.choice([True, False], p=additional_words_probabilities)

        additional_word = ''
        if is_additional_alphabet_word:
            additional_word_random_index = self.random_provider.randint(0, additional_alphabet_words_len - 1)
            additional_word = additional_alphabet_words[additional_word_random_index]
        else:
            additional_numbers_len = EOS_NAME_LENGTH - len(base_word)
            for _ in range(additional_numbers_len):
                additional_char = str(self.random_provider.randint(1, 5))
                additional_word += additional_char

        random_name = base_word + additional_word
        return random_name

    def __get_probabilities_len_base_word(self) -> list:
        """
        Get probabilities list of the base word length based on word frequency.

        :return: probabilities list
        """
        probabilities = []
        base_dict_len = 0

        for key in self.__base_dict:
            word_count = len(self.__base_dict.get(key))
            base_dict_len += word_count
            probabilities.append(word_count)

        probabilities = [word_count / base_dict_len for word_count in probabilities]

        return probabilities

    def __get_probability_alphabet_additional_word(self, additional_alphabet_words_len) -> int:
        """
        Get probability of the additional word alphabet.

        Get probability of the additional word alphabet based on `numbers_probabilities`
        and `additional_alphabet_words_len`.

        :return: probabilities list
        """
        additional_words_probability = 0
        if additional_alphabet_words_len:
            additional_words_probability = 1 - self.numbers_probabilities

        return additional_words_probability

    def __repr__(self):
        """
        Debug `repr` method.

        :return: `RandomNameGenerator` object state
        """
        return f'<RandomNameGenerator({self.seed_data_path}, {self.numbers_probabilities}, {self.random_provider})>'



"""
Provide an implementation of the FastRandomChoice interface.
"""
import random

import numpy as np

from eos_name_generator.utils.fast_random_choice.interfaces import FastRandomChoiceInterface


class FastRandomChoice(random.Random, FastRandomChoiceInterface):
    """
    Implementation of the FastRandomChoice.
    """

    def choice(self, seq, p):
        """
        Choose a random element from a non-empty sequence with probabilities list.

        :param seq: non-empty sequence.
        :param p: probabilities according to sequence.
        :return: random element from sequence
        """
        sequence_len = len(seq)
        if not sequence_len:
            raise IndexError('Cannot choose from an empty sequence')

        if sequence_len != len(p):
            raise ValueError('`seq` and `p` must have same size')

        is_probabilities_positive = all(i >= 0 for i in p)
        if not is_probabilities_positive:
            raise ValueError('Probabilities are not non-negative')

        probabilities_sum = sum(p)
        if 0.99 > probabilities_sum < 1.01:
            raise ValueError('Probabilities do not sum to 1')

        element_index = self.multidimensional_shifting(probabilities=p)
        random_element = seq[element_index]

        return random_element

    @staticmethod
    def multidimensional_shifting(probabilities: list) -> int:
        """
        Get the most probable element from probabilities sequence.

        This method is a direct replacement for slow cycle python loops.
        :param probabilities: probabilities list.
        :return: probability index.
        """
        # replicate probabilities as many times
        replicated_probabilities = np.tile(probabilities, (1, 1))
        # get random shifting numbers & scale them correctly
        random_shifts = np.random.random(replicated_probabilities.shape)
        random_shifts /= random_shifts.sum(axis=1)[:, np.newaxis]
        # shift by numbers & find largest (by finding the smallest of the negative)
        shifted_probabilities = random_shifts - replicated_probabilities
        index_list = np.argpartition(shifted_probabilities, 1, axis=1)[:, :1]

        index = int(index_list)
        return index



"""
Provide implementation of the FastRandomChoice interfaces.
"""
from abc import (
    ABC,
    abstractmethod,
)


class FastRandomChoiceInterface(ABC):
    """
    Provide implementation of the FastRandomChoice interfaces.
    """

    @abstractmethod
    def choice(self, seq, p):
        """
        Choose a random element from a non-empty sequence with probabilities list.

        :param seq: non-empty sequence.
        :param p: probabilities according to sequence.
        :return: random element from sequence
        """



"""
Provide implementation of the DataReader.
"""
from collections import defaultdict

from eos_name_generator.constants import EOS_NAME_LENGTH
from eos_name_generator.errors import ValidationDataError


class DataReader:
    """
    Implementation of the DataReader.
    """

    def __init__(self, data_path):
        self.data_path = data_path

    def get_dictionary_by_word_len(self) -> defaultdict:
        """
        Read data from `seed_data_path`.

        Read data from `seed_data_path` and transform it into `dictionary` object
        where the key is the word length.
        """
        with open(self.data_path) as f:
            data = f.read().splitlines()

        self.__validate_data(data)
        data_dictionary_by_word_len = defaultdict(list)
        for word in data:
            word_len = len(word)
            data_dictionary_by_word_len[word_len].append(word)

        return data_dictionary_by_word_len

    @staticmethod
    def __validate_data(data: list):
        """
        Seed data validation to generate the correct `eos` name.

        :param data: data to be validated
        """
        for word in data:
            is_valid_word_len = (len(word) <= EOS_NAME_LENGTH)
            is_valid_word = word.isalpha() and word.islower() and is_valid_word_len

            if not is_valid_word:
                raise ValidationDataError("Data contains invalid characters or does not match the name length error")



"""
Setup the package.
"""
from setuptools import find_packages, setup

with open('README.md') as read_me:
    long_description = read_me.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    version='0.1.0',
    name='eos-name-generator',
    description='Python package for generation random name which suits `EOS` name conversations',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Alladin9393/name-generator',
    license='MIT',
    author='Alladin9393',
    author_email='ember.toon@protonmail.com',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
